// ----------------------------------------------------------
//  ""
// :  ..
// ()   , 2013
// ----------------------------------------------------------

#include "Experiment.h"

#include <sstream>
#include <iomanip>
#include <fstream>

#include "Timer.h"

#include "ImageLoader.h"
#include "Fresnel.h"
#include "DHI.h"
#include "Convolution.h"
#include "Experiment.h"
#include "MatrixHelper.h"
#include "SConvert.h"
#include "Constants.h"

namespace Holo
{

void SaveBinary( const std::string& filename, const double** data, int size )
{
	std::ofstream outputStream(filename);
	for (int x = 0; x < size; x++)
	{
		for (int y = 0; y < size; y++)
		{
			outputStream.write(reinterpret_cast<const char*>(&data[x][y]), sizeof(data[x][y]));
		}
	}
	outputStream.close();
}

Experiment::Experiment() :
	m_type(ExperimentType::Reconstruction),
	m_recType(ReconstructionType::Fresnel),
	m_clipType(ImageLoader::ClipType::Center),
	m_preferredSize(0)
{

}

bool Experiment::Load(const std::string& filename)
{
	m_reconstructionParams.clear();

	pugi::xml_document doc;
	pugi::xml_parse_result result = doc.load_file(filename.c_str());
	if (!result) return false;

	pugi::xml_node experiment_root = doc.child("Experiment");
	if (experiment_root.empty()) return false;

	pugi::xml_attribute experiment_type_attr = experiment_root.attribute("type");
	if (experiment_type_attr.empty()) return false;
	std::string exptype = experiment_type_attr.value();
	if (exptype == "Reconstruction") m_type = ExperimentType::Reconstruction;
	else if (exptype == "DHI") m_type = ExperimentType::DHI;

	pugi::xml_attribute method_attr = experiment_root.attribute("method");
	if (method_attr.empty()) return false;
	std::string methodtype = method_attr.value();
	if (methodtype == "Fresnel") m_recType = ReconstructionType::Fresnel;
	else if (methodtype == "Convolution") m_recType = ReconstructionType::Convolution;

	pugi::xml_node hologram_node = experiment_root.child("Hologram");
	if (hologram_node.empty()) return false;
	pugi::xml_attribute hologram_attr = hologram_node.attribute("value");
	if (hologram_attr.empty()) return false;
	if (m_type == ExperimentType::DHI)
	{
		pugi::xml_attribute hologram2_attr = hologram_node.attribute("value2");
		if (hologram2_attr.empty()) return false;
		m_hologram2Path = hologram2_attr.value();
	}
	pugi::xml_attribute preferredSize_attr = hologram_node.attribute("preferredSize");
	if (preferredSize_attr.empty()) return false;
	pugi::xml_attribute clipType_attr = hologram_node.attribute("clipType");
	if (clipType_attr.empty()) return false;

	m_hologramPath = hologram_attr.value();
	std::string cliptype = clipType_attr.value();
	m_clipType = ImageLoader::ClipType::Center;
	if (cliptype == "Center") m_clipType = ImageLoader::ClipType::Center;
	else if (cliptype == "Left") m_clipType = ImageLoader::ClipType::Left;
	else if  (cliptype == "Right") m_clipType = ImageLoader::ClipType::Right;
	m_preferredSize = preferredSize_attr.as_int();

	pugi::xml_node wavelength_node = experiment_root.child("Wavelength");
	if (wavelength_node.empty()) return false;
	std::list<double> wavelength_values = GetValues(wavelength_node);
	if (wavelength_values.empty()) return false;

	pugi::xml_node angle_node = experiment_root.child("Angle");
	if (angle_node.empty()) return false;
	std::list<double> angle_values = GetValues(angle_node);
	if (angle_values.empty()) return false;

	pugi::xml_node zdist_node = experiment_root.child("ZDistance");
	if (zdist_node.empty()) return false;
	std::list<double> zdist_values = GetValues(zdist_node);
	if (zdist_values.empty()) return false;

	pugi::xml_node ccdpd_node = experiment_root.child("CcdPixelDistance");
	if (ccdpd_node.empty()) return false;
	std::list<double> ccdpd_values = GetValues(ccdpd_node);
	if (ccdpd_values.empty()) return false;

	m_reconstructionParams.reserve(wavelength_values.size() * angle_values.size() * zdist_values.size() * ccdpd_values.size());
	for (std::list<double>::iterator w = wavelength_values.begin(); w != wavelength_values.end(); w++)
	{
		for (std::list<double>::iterator a = angle_values.begin(); a != angle_values.end(); a++)
		{
			for (std::list<double>::iterator z = zdist_values.begin(); z != zdist_values.end(); z++)
			{
				for (std::list<double>::iterator c = ccdpd_values.begin(); c != ccdpd_values.end(); c++)
				{
					ReconstructionParams params;
					params.waveLength = *w;
					params.angle = *a;
					params.zDistance = *z;
					params.ccdPixelDistance = *c;

					m_reconstructionParams.push_back(params);
				}
			}
		}
	}

	pugi::xml_node output_node = experiment_root.child("Output");
	if (output_node.empty()) return false;
	pugi::xml_attribute output_norm_attr = output_node.attribute("normalize");
	if (output_norm_attr.empty()) return false;
	pugi::xml_attribute output_brightness_attr = output_node.attribute("brightness");
	if (output_brightness_attr.empty()) return false;

	m_outputParams.normalize = output_norm_attr.as_bool();
	m_outputParams.brigthness = output_brightness_attr.as_double();

	return true;
}

std::list<double> Experiment::GetValues(const pugi::xml_node& node)
{
	std::list<double> lst;
	pugi::xml_node single_node = node.child("Single");
	if (!single_node.empty())
	{
		pugi::xml_attribute attr = single_node.attribute("value");
		if (attr.empty()) return lst;
		lst.push_back(attr.as_double());
		return lst;
	}
	else
	{
		pugi::xml_node range_node = node.child("Range");
		if (range_node.empty()) return lst;
		pugi::xml_attribute min_attr = range_node.attribute("minVal");
		pugi::xml_attribute max_attr = range_node.attribute("maxVal");
		pugi::xml_attribute step_attr = range_node.attribute("step");
		if (min_attr.empty() || max_attr.empty() || step_attr.empty()) return lst;
		double min_val = min_attr.as_double();
		double max_val = max_attr.as_double();
		double step = step_attr.as_double();
		for (double s = min_val; s <= max_val; s += step) lst.push_back(s);
		return lst;
	}
}

void Experiment::SaveResults(const std::string& filename, double fullTime, double avgTime)
{
	pugi::xml_document doc;
	pugi::xml_node results_node = doc.append_child("Results");
	results_node.append_attribute("hologramsCount") = SConvert::ToString((int)m_reconstructionParams.size()).c_str();
	results_node.append_attribute("fullRTime") = SConvert::ToString((float)fullTime).c_str();
	results_node.append_attribute("averageTime") = SConvert::ToString((float)avgTime).c_str();

	for(size_t i = 0; i < m_reconstructionParams.size(); i++)
	{
		pugi::xml_node parameters_node = results_node.append_child("Parameters");
		parameters_node.append_attribute("index") = SConvert::ToString((int)i).c_str();

		parameters_node.append_child("WaveLength").append_attribute("value") = SConvert::ToString((float)m_reconstructionParams[i].waveLength).c_str();
		parameters_node.append_child("Angle").append_attribute("value") = SConvert::ToString((float)m_reconstructionParams[i].angle).c_str();
		parameters_node.append_child("ZDistance").append_attribute("value") = SConvert::ToString((float)m_reconstructionParams[i].zDistance).c_str();
		parameters_node.append_child("CcdPixelDistance").append_attribute("value") = SConvert::ToString((float)m_reconstructionParams[i].ccdPixelDistance).c_str();
	}

	doc.save_file(filename.c_str());
}

std::string GetFileNameFromPath(const std::string& path)
{
	char drive[_MAX_DRIVE];
	char dir[_MAX_DIR];
	char fname[_MAX_FNAME];
	char ext[_MAX_EXT];
	_splitpath_s(path.c_str(), drive, dir, fname, ext);
	return std::string(fname) + ext;
}

void Experiment::Undertake(const std::string& filename, IExperimentListener* listener)
{
	time_t t = time(0);
	tm now;
    localtime_s(&now, &t);

	std::stringstream folderStream;
	folderStream.clear();
	folderStream << std::setfill('0');
	folderStream << (now.tm_year + 1900) << '_' << std::setw(2) << (now.tm_mon + 1) << '_' << std::setw(2) << now.tm_mday << '_' << std::setw(2) << now.tm_hour << '_' << std::setw(2) << now.tm_min << '_' << std::setw(2) << now.tm_sec;
	std::string resultFolder = folderStream.str();
	CreateDirectory(resultFolder.c_str(), NULL);

	if (!Load(filename))
	{
		if (listener != 0) listener->OnError("Can't load a experiment");
		return;
	}
	
	if (listener != 0) listener->OnStart(filename, now, m_type, m_recType);

	m_timerFull.Clear();
	m_timerFull.Start();

	double tm = 0;
	if (m_type == ExperimentType::Reconstruction)
	{
		double** data = 0;
		int size = 0;
		LoadImage(listener, m_hologramPath, data, size);

		double** intensity = new double*[size];
		for (int i = 0; i < size; i++) intensity[i] = new double[size];
		double** phase = new double*[size];
		for (int i = 0; i < size; i++) phase[i] = new double[size];

		for(size_t i = 0; i < m_reconstructionParams.size(); i++)
		{
			m_timer.Clear();
			m_timer.Start();
			Step(listener, (int)i, (const double**)data, size, intensity, phase, m_reconstructionParams[i], m_outputParams, resultFolder);
			m_timer.Stop();
			double t = m_timer.GetTime();
			tm += t;
			double average = tm / (i + 1);
			if (listener != 0) listener->OnStep(i + 1, (int)m_reconstructionParams.size(), t, average);
		}
		tm /= (double)m_reconstructionParams.size();

		RELEASE_MATRIX(data, size);
		RELEASE_MATRIX(intensity, size);
		RELEASE_MATRIX(phase, size);
	}
	else if (m_type == ExperimentType::DHI)
	{
		double** data1 = 0;
		double** data2 = 0;
		int size = 0;
		LoadImage(listener, m_hologramPath, data1, size);
		LoadImage(listener, m_hologram2Path, data2, size);

		double** intensity1 = new double*[size];
		for (int i = 0; i < size; i++) intensity1[i] = new double[size];
		double** intensity2 = new double*[size];
		for (int i = 0; i < size; i++) intensity2[i] = new double[size];
		double** phaseDiff = new double*[size];
		for (int i = 0; i < size; i++) phaseDiff[i] = new double[size];
		double** phaseUnwrapped = new double*[size];
		for (int i = 0; i < size; i++) phaseUnwrapped[i] = new double[size];

		for(size_t i = 0; i < m_reconstructionParams.size(); i++)
		{
			m_timer.Clear();
			m_timer.Start();
			StepDHI(listener, (int)i, (const double**)data1, (const double**)data2, size, intensity1, intensity2, phaseDiff, phaseUnwrapped, m_reconstructionParams[i], m_outputParams, resultFolder);
			m_timer.Stop();
			double t = m_timer.GetTime();
			tm += t;
			double average = tm / (i + 1);
			if (listener != 0) listener->OnStep(i + 1, (int)m_reconstructionParams.size(), t, average);
		}
		tm /= (double)m_reconstructionParams.size();

		RELEASE_MATRIX(data1, size);
		RELEASE_MATRIX(data2, size);
		RELEASE_MATRIX(intensity1, size);
		RELEASE_MATRIX(intensity2, size);
		RELEASE_MATRIX(phaseDiff, size);
		RELEASE_MATRIX(phaseUnwrapped, size);
	}

	m_timerFull.Stop();

	SaveResults(resultFolder + "/results.xml", m_timerFull.GetTime(), tm);

	CreateDirectory((resultFolder + "/params").c_str(), NULL);
	CopyFile(m_hologramPath.c_str(), (resultFolder + "/params/" + GetFileNameFromPath(m_hologramPath)).c_str(), 0);
	CopyFile(filename.c_str(), (resultFolder + "/params/" + GetFileNameFromPath(filename)).c_str(), 0);

	if (listener != 0) listener->OnFinish();
}

void Experiment::Step(IExperimentListener* listener, int stepIndex, const double** hologram, int sizeInPixel,  
					  double**& intensity, double**& phase, ReconstructionParams& recParams, OutputParams& outputParams,
					  const std::string& resultFolder)
{
	if (m_recType == ReconstructionType::Fresnel)
	{
		Holo::Fresnel::Reconstruct((const double**)hologram, sizeInPixel, recParams.waveLength, recParams.angle, recParams.zDistance, recParams.ccdPixelDistance, intensity, phase);
	}
	else if (m_recType == ReconstructionType::Convolution)
	{
		Holo::Convolution::Reconstruct((const double**)hologram, sizeInPixel, recParams.waveLength, recParams.angle, recParams.zDistance, recParams.ccdPixelDistance, intensity, phase);
	}
	
	std::stringstream outFilename;
	outFilename << std::setfill('0');
	outFilename << resultFolder << "/result_" << std::setw(2) << stepIndex;
	std::stringstream errorStream;
	if (!Holo::ImageLoader::Instance().Save(outFilename.str(), intensity, sizeInPixel, outputParams.normalize, outputParams.brigthness, errorStream))
	{
		if (listener != 0) listener->OnError(errorStream.str());
	}
}

void Experiment::StepDHI(IExperimentListener* listener, int stepIndex, const double** hologram1, const double** hologram2, int sizeInPixel, 
						 double**& intensity1, double**& intensity2, double**& phaseDiff, double**& phaseUnwrapped, ReconstructionParams& recParams, OutputParams& outputParams,
						 const std::string& resultFolder)
{
	Holo::DHI::ReconstructionType rtype = Holo::DHI::ReconstructionType::Fresnel;
	if (m_recType == ReconstructionType::Fresnel) rtype = Holo::DHI::ReconstructionType::Fresnel;
	else if (m_recType == ReconstructionType::Convolution) rtype = Holo::DHI::ReconstructionType::Convolution;

	Holo::DHI::Reconstruct(rtype, hologram1, hologram2, sizeInPixel, recParams.waveLength, recParams.angle, recParams.zDistance, recParams.ccdPixelDistance, intensity1, intensity2, phaseDiff);

	std::stringstream outFilename;
	outFilename << std::setfill('0');
	outFilename << resultFolder << "/result_" << std::setw(2) << stepIndex;
	std::stringstream errorStream;
	if (!Holo::ImageLoader::Instance().Save(outFilename.str() + "_intensity1", intensity1, sizeInPixel, outputParams.normalize, outputParams.brigthness, errorStream))
	{
		if (listener != 0) listener->OnError(errorStream.str());
	}
	if (!Holo::ImageLoader::Instance().Save(outFilename.str() + "_intensity2", intensity2, sizeInPixel, outputParams.normalize, outputParams.brigthness, errorStream))
	{
		if (listener != 0) listener->OnError(errorStream.str());
	}
	if (!Holo::ImageLoader::Instance().Save(outFilename.str() + "_phaseDiff", phaseDiff, sizeInPixel, true, 1.0, errorStream))
	{
		if (listener != 0) listener->OnError(errorStream.str());
	}
}

void Experiment::LoadImage( IExperimentListener* listener, const std::string& filename, double**& data, int& size )
{
	std::stringstream errorStream;
	if (m_preferredSize != 0)
	{
		size = m_preferredSize;
		if (!Holo::ImageLoader::Instance().Load(filename, data, size, m_clipType, errorStream))
		{
			if (listener != 0) listener->OnError(errorStream.str());
			return;
		}
	}
	else
	{
		if (!Holo::ImageLoader::Instance().Load(filename, data, size, errorStream))
		{
			if (listener != 0) listener->OnError(errorStream.str());
			return;
		}
	}
}

}