// ----------------------------------------------------------
//  ""
// :  ..
// ()   , 2013
// ----------------------------------------------------------

#include "ImageLoader.h"

#include <math.h>

namespace Holo
{

ImageLoader::ImageLoader()
{
}

ImageLoader::~ImageLoader()
{
}

ImageLoader& ImageLoader::Instance()
{
	static ImageLoader loader;
	return loader;
}

FIBITMAP* ImageLoader::GenericLoader(const std::string& filename)
{	
	FREE_IMAGE_FORMAT fif = FreeImage_GetFileType(filename.c_str(), 0);
	if(fif == FIF_UNKNOWN)
	{
		fif = FreeImage_GetFIFFromFilename(filename.c_str());
	}
	
	if((fif != FIF_UNKNOWN) && FreeImage_FIFSupportsReading(fif))
	{
		FIBITMAP *dib = FreeImage_Load(fif, filename.c_str(), 0);
		return dib;
	}
	
	return NULL;
}

bool ImageLoader::Load(const std::string& filename, double**& data, int& size, std::stringstream& errorStr)
{
	FIBITMAP* p_image = GenericLoader(filename);
	if (p_image == 0)
	{
		errorStr << "The image couldn't be loaded" << std::endl;
		return false;
	}
	
	int iw = FreeImage_GetWidth(p_image);
	int ih = FreeImage_GetHeight(p_image);
	if (iw != ih)
	{
		errorStr << "The image must be square" << std::endl;
		FreeImage_Unload(p_image);
		return false;
	}
	size = iw;

	bool result = Load(p_image, data, size, 0, 0, errorStr);
	FreeImage_Unload(p_image);
	return result;
}

bool ImageLoader::Load(const std::string& filename, double**& data, int preferredSize, int xoffset, int yoffset, std::stringstream& errorStr)
{
	FIBITMAP* p_image = GenericLoader(filename);
	if (p_image == 0)
	{
		errorStr << "The image couldn't be loaded" << std::endl;
		return false;
	}

	if (preferredSize == 0)
	{
		int iw = FreeImage_GetWidth(p_image);
		int ih = FreeImage_GetHeight(p_image);
		if (iw != ih)
		{
			errorStr << "The image must be square" << std::endl;
			FreeImage_Unload(p_image);
			return false;
		}
		preferredSize = iw;
	}
	
	bool result = Load(p_image, data, preferredSize, xoffset, yoffset, errorStr);
	FreeImage_Unload(p_image);
	return result;
}

bool ImageLoader::Load(const std::string& filename, double**& data, int preferredSize, ClipType clipType, std::stringstream& errorStr)
{
	FIBITMAP* p_image = GenericLoader(filename);
	if (p_image == 0)
	{
		errorStr << "The image couldn't be loaded" << std::endl;
		return false;
	}
	int iw = FreeImage_GetWidth(p_image);
	int ih = FreeImage_GetHeight(p_image);

	int yoffset = (ih - preferredSize) / 2;
	int xoffset = 0;
	if (clipType == ClipType::Left) xoffset = 0;
	else if (clipType == ClipType::Center) xoffset = (iw - preferredSize) / 2;
	else if (clipType == ClipType::Right) xoffset = iw - preferredSize;

	bool result = Load(p_image, data, preferredSize, xoffset, yoffset, errorStr);
	FreeImage_Unload(p_image);
	return result;
}

bool ImageLoader::Load(FIBITMAP* image, double**& data, int preferredSize, int xoffset, int yoffset, std::stringstream& errorStr)
{
	if (!image) return false;

	int iw = FreeImage_GetWidth(image);
	int ih = FreeImage_GetHeight(image);
	if (iw == 1 || ih == 1) 
	{
		errorStr << "The image size couldn't be 1" << std::endl;
		return false;
	}

	data = new double*[preferredSize];
	for (int i = 0; i < preferredSize; i++) data[i] = new double[preferredSize];

	for (int x = 0; x < preferredSize; x++)
	{
		for (int y = 0; y < preferredSize; y++)
		{
			int yy = y + xoffset;
			int xx = x + yoffset;
			if (yy > iw || xx > ih || yy < 0 || xx < 0) { data[x][y] = 0; continue; }

			RGBQUAD pixel;
			FreeImage_GetPixelColor(image, yy, xx, &pixel);
			data[x][y] = (double)pixel.rgbRed;
		}
	}

	return true;
}

bool ImageLoader::Save(const std::string& filename, double** data, int size, bool normalize, double brightness, std::stringstream& errorStr)
{
	FIBITMAP* image = FreeImage_Allocate(size, size, 24);

	double minValue = FLT_MAX;
	double maxValue = FLT_MIN;
	if (normalize)
	{
		for (int i = 0; i < size; i++)
		{
			for (int j = 0; j < size; j++)
			{
				if (data[i][j] < minValue) minValue = data[i][j];
				if (data[i][j] > maxValue) maxValue = data[i][j];
			}
		}
	}

	RGBQUAD color;
	for (int i = 0; i < size; i++)
	{
		for (int j = 0; j < size; j++)
		{
			double val = 0;
			if (normalize) val = 255.0 * brightness * (data[i][j] - minValue) / (maxValue - minValue);
			else val = brightness * data[i][j];
			if (val < 0) val = 0;
			if (val > 255) val = 255;
			
			color.rgbRed = (unsigned char)val;
			color.rgbGreen = (unsigned char)val;
			color.rgbBlue = (unsigned char)val;
			FreeImage_SetPixelColor(image, j, i, &color);
		}
	}

	bool result = FreeImage_Save(FIF_PNG, image, (filename + ".png").c_str(), 0) != FALSE;
	FreeImage_Unload(image);

	return result;
}

}